///////////////////////////////////////////////////////////////////
// Alpine M-Bus interface to RS232
// (c) 2003 Joerg Hohensohn, allowed for non-commercial use only
//                           (feedback appreciated)
//
// This translates the M-Bus to RS232 and back (listen and send)
// For more info see http://joerg.hohensohn.bei.t-online.de/mbus
///////////////////////////////////////////////////////////////////

#include <inttypes.h>      // common scalar types
#include <avr/io.h>        // for device register definitions
#include <avr/interrupt.h> // for interrupt enable
#include <avr/sleep.h>     // for power-save idle sleep
#include <avr/wdt.h>       // for watchdog, used to prevent deadlocks in interrupts
#include <avr/eeprom.h>    // EEPROM access
#include "main.h"         // look for definitions here
#include "cdc.h"

// m_bus buffer size definitions
#define M_BUS_RX_BUFFER_SIZE 16     // 2,4,8,16,32,64,128 or 256 bytes
#define M_BUS_TX_BUFFER_SIZE 256     // 2,4,8,16,32,64,128 or 256 bytes

#define M_BUS_RX_BUFFER_MASK (M_BUS_RX_BUFFER_SIZE - 1)
#if (M_BUS_RX_BUFFER_SIZE & M_BUS_RX_BUFFER_MASK)
	#error M_BUS RX buffer size is not a power of 2
#endif

#define M_BUS_TX_BUFFER_MASK (M_BUS_TX_BUFFER_SIZE - 1)
#if (M_BUS_TX_BUFFER_SIZE & M_BUS_TX_BUFFER_MASK)
	#error M_BUS TX buffer size is not a power of 2
#endif

static uint8_t M_BUS_RxBuf[M_BUS_RX_BUFFER_SIZE];
static volatile uint8_t M_BUS_RxHead;
static volatile uint8_t M_BUS_RxTail;

static uint8_t M_BUS_TxBuf[M_BUS_TX_BUFFER_SIZE];
static volatile uint8_t M_BUS_TxHead;
static volatile uint8_t M_BUS_TxTail;


static struct
{	// all the information for the receive state
	volatile enum
	{
		eWait, // waiting for packet start
		eHigh, // rising edge has been seen
		eLow,  // falling edge has been seen
	} state;
	char a4bits[4]; // received bits as chars
	uint8_t uBits; // # of received bits
} gRcv; 

static struct
{	// all the information for the transmit state
	uint8_t uBits; // # of sent bits
	uint8_t uCurrent; // processed hex nibble
	enum
	{
		eStart, // before sending a bit
		eLow0,  // short pulse of a '0' is sent
		eLow1,  // long pulse of a '1' is sent
		eEnd,   // end of sequence
	} state;
} gTrans; 

volatile static uint8_t half_sec_flag = 0;
volatile static uint8_t one_sec_flag = 0;
volatile static uint8_t power_off_flag = 0;
volatile static uint8_t power_off_timeout_flag = 0;
volatile static uint8_t i2c_addr, i2c_data, twi_in_use;



/*
******************************
* MAIN
*
******************************
*/

int main (void)
{
	wdt_reset(); // reset watchdog before enabling it
//	wdt_enable(WDTO_15MS); // enable with safe interval, larger than timer1 timeout

    init_io(); // setup timers and ports

	M_BUS_RxTail = 0;
	M_BUS_RxHead = 0;
	M_BUS_TxTail = 0;
	M_BUS_TxHead = 0;

	cdc_init();

	// setup the states, only the necessary
	gRcv.state = eWait;
	gTrans.uBits = 0;
	
    sei(); // enable interrupts

    for (;;)
    {
//    	sleep_mode(); // idle mode by default
    	wdt_reset(); // if nothing happens, timer1 keeps waking us up
    	
    	// check if there is a command to be sent
    	if (MbusSearchInRxBuffer(0xF0)) // new command in buffer
    	{	// start the transmission
			cdc_process_command();
    	}

    	if (!(TIMSK & _BV(TOIE0)) // not already sending
			&& MbusSearchInTxBuffer(0xF0) // new command in buffer
    		&& gRcv.state == eWait // not receiving
			)
    	{	// start the transmission

			gTrans.state = eStart;
			TCCR0  = PRESCALE_256; // slow prescaling while sending
			TCNT0 = 0; // reset timer because ISR only offsets to it
			TIMSK |= _BV(TOIE0); // start the output handler with timer0
		}

		if (one_sec_flag && !MbusSearchInTxBuffer(0xF0))
		{
			one_sec_flag = 0;
			cdc_update_status();
		}

		if (power_off_flag) 
		{
			m32_send_command(m32_STOP); //send stop command 0.5sec after power line falls down
			power_off_flag = 0;
		}

		if (power_off_timeout_flag) 
		{
			m32_send_command(m32_SAVE); //send save and stop command 1 min after power line falls down
			
			uint8_t i=6;
			while(i){
				if (half_sec_flag)
				{
					half_sec_flag = 0;
					i--;
					if (m32_read_status() == m32_STOPPED) break;  //waiting for stopped status
		
				}
			}

			PORTD &= ~(_BV(PIN_POWER_RELAY)); //power off

			while (!(PIND & _BV(PIN_POWER_ON))); 

			power_off_timeout_flag = 0;
			PORTD |= _BV(PIN_POWER_RELAY);
			
		}

    }

    return (0);
}


/*
******************************
* 
*
******************************
*/


// helper function for main(): setup timers and pins
void init_io (void)
{
	// set my outputs, all but PB2 are for debugging / error signaling
	DDRB =  _BV(PIN_MBUS_OUT); // output and debug pins

	DDRD = _BV(PIN_POWER_RELAY);
	PORTD = _BV(PIN_POWER_RELAY); //power on

	// timer settings: prescale 256 = 28,8kHz
	TCCR0  = PRESCALE_1; // fast prescale that would immediately generate interrups
	TCNT0 = -1; // next interrupt will be pending immediately, but is masked
	TCCR1B = _BV(ICNC1) | _BV(WGM12) | PRESCALE_256; // noise filter, reset on match, prescale
	OCR1AH = 0; // we use only the lower part, but have to write this first
	OCR1AL = BIT_TIMEOUT; // have to complete a bit within this time

	TCCR2 = 7; //run sys timer

    // enable capture and compare match interrupt for timer 1 and compare match for timer 2
	TIMSK = _BV(TICIE1) | _BV(OCIE1A) | _BV(TOIE2);
	
	//TWI config in master mode
	TWBR = 0x07;
	TWSR = 0x00;
}


// utility function: convert a number to a hex char
char int2hex(uint8_t n)
{
	if (n<10)
		return ('0' + n);
	else if (n<16)
		return ('A' + n - 10);

	return 'X'; // out of range
}

// utility function: convert a hex char to a number
uint8_t hex2int(char c)
{
	if (c < '0')
		return 0xFF; // illegal hex digit
	else if (c <= '9')
		return (c - '0');
	else if (c < 'A')
		return 0xFF; // illegal hex digit
	else if (c <= 'F')
		return (c - 'A' + 10);
	else if (c < 'a')
		return 0xFF; // illegal hex digit
	else if (c <= 'f')
		return (c - 'a' + 10);

	return 0xFF; // default
}

/*
******************************
* m32 control functions
*
******************************
*/

void m32_send_command(uint8_t command)
{
	while (twi_in_use);
	twi_in_use = 0xFF;
	i2c_addr = m32_addr_w;
	i2c_data = command;
	TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN)|(1<<TWIE);
}

uint8_t m32_read_status(void)
{
	while (twi_in_use);
	twi_in_use = 0xFF;
	i2c_addr = m32_addr_r;
	TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN)|(1<<TWIE);
	while (twi_in_use);
	return i2c_data;
}

/*
******************************
* buffer functions
*
******************************
*/

uint8_t MbusRxReadByte(void)
{
	uint8_t tmptail;
	// calculate buffer index
	tmptail = (M_BUS_RxTail + 1) & M_BUS_RX_BUFFER_MASK; // wait for free space in buffer

	M_BUS_RxTail = tmptail; // store new index
	return M_BUS_RxBuf[tmptail]; // return data from buffer
}


void MbusRxWriteByte(uint8_t data)
{
	uint8_t tmphead;
	// calculate buffer index
	tmphead = (M_BUS_RxHead + 1) & M_BUS_RX_BUFFER_MASK; // wait for free space in buffer

	M_BUS_RxBuf[tmphead] = data; // store data in buffer
	M_BUS_RxHead = tmphead; // store new index
}


uint8_t MbusTxReadByte(void)
{
	uint8_t tmptail;
	// calculate buffer index
	tmptail = (M_BUS_TxTail + 1) & M_BUS_TX_BUFFER_MASK; // wait for free space in buffer

	M_BUS_TxTail = tmptail; // store new index
	return M_BUS_TxBuf[tmptail]; // return data from buffer
}

void MbusTxWriteByte(uint8_t data)
{
	uint8_t tmphead;
	// calculate buffer index
	tmphead = (M_BUS_TxHead + 1) & M_BUS_TX_BUFFER_MASK; // wait for free space in buffer

	M_BUS_TxBuf[tmphead] = data; // store data in buffer
	M_BUS_TxHead = tmphead; // store new index
}


// return 0 (FALSE) if the receive buffer is empty
uint8_t MbusDataInRxBuffer(void)
{
	return (M_BUS_RxHead - M_BUS_RxTail) & M_BUS_RX_BUFFER_MASK;
}

// search for a specific char in buffer, return position (1-based, 0 if not found)
uint8_t MbusSearchInRxBuffer(uint8_t key)
{
	uint8_t i;
	uint8_t level = MbusDataInRxBuffer();

	for (i=1; i<=level; i++)
	{
		if (M_BUS_RxBuf[(M_BUS_RxTail + i) & M_BUS_RX_BUFFER_MASK] == key)
			return i; // found
	}
	return 0; // not found
}

// return 0 (FALSE) if the receive buffer is empty
uint8_t MbusDataInTxBuffer(void)
{
	return (M_BUS_TxHead - M_BUS_TxTail) & M_BUS_TX_BUFFER_MASK;
}

// search for a specific char in buffer, return position (1-based, 0 if not found)
uint8_t MbusSearchInTxBuffer(uint8_t key)
{
	uint8_t i;
	uint8_t level = MbusDataInTxBuffer();

	for (i=1; i<=level; i++)
	{
		if (M_BUS_TxBuf[(M_BUS_TxTail + i) & M_BUS_TX_BUFFER_MASK] == key)
			return i; // found
	}
	return 0; // not found
}


/*
******************************
* ISR vectors
*
******************************
*/

// timer 0 overflow, used for sending
ISR(TIMER0_OVF_vect)
{
	switch(gTrans.state)
	{
	case eStart: // start of a bit, or maybe the packet
		if (gTrans.uBits % 4 == 0)
		{	// need a new hex nibble
			uint8_t fetched;
				fetched = MbusTxReadByte();
				gTrans.uCurrent = fetched;
			
			if (fetched == 0xF0)
			{	// done with this line (packet)
				TCNT0 -= SEND_SPACE; // space til the next transmision can start
				gTrans.state = eEnd;
				break; // exit
			}
		}
		PORTB |= _BV(PIN_MBUS_OUT); // pull the line low

		if (gTrans.uCurrent & (1 << (3 - (gTrans.uBits % 4))))
		{	// 1
			gTrans.state = eLow1;
			TCNT0 -= SEND_ONE_TIME; // next edge for the long pulse
		}
		else
		{	// 0
			gTrans.state = eLow0;
			TCNT0 -= SEND_ZERO_TIME; // next edge for the short pulse
		}
		gTrans.uBits++; // next bit
		break;

	case eLow0:
		PORTB &= ~_BV(PIN_MBUS_OUT); // release the line
		gTrans.state = eStart;
		TCNT0 -= (SEND_BIT_TIME - SEND_ZERO_TIME); // next edge
		break;
	
	case eLow1:
		PORTB &= ~_BV(PIN_MBUS_OUT); // release the line
		gTrans.state = eStart;
		TCNT0 -= (SEND_BIT_TIME - SEND_ONE_TIME); // next edge
		break;
		
	case eEnd:
		TIMSK &= ~_BV(TOIE0); // stop timer0 interrupts
		TCCR0  = PRESCALE_1; // set to "immediate interrupt" mode again
		TCNT0 = -1; // next interrupt will be pending immediately, but is masked
		gTrans.state = eStart;
		gTrans.uBits = 0; // reset the bit counter again
		break;
	}
}

// edge detection interrupt, the heart of receiving
ISR(TIMER1_CAPT_vect)
{	
	char outchar;
	
	switch(gRcv.state)
	{
	case eWait: // a packet is starting
		gRcv.uBits = 0;
		// no break, fall through
	case eHigh: // high phase between bits has ended, start of low pulse
		// could check the remain high time to verify bit, but won't work for the last (timed out)
		TCNT1H = 0; // reset the timer, high byte first
		TCNT1L = 0;
		TIFR |= _BV(OCF1A); // clear timeout pending, to be shure
	
		gRcv.state = eLow;
		TCCR1B |= _BV(ICES1); // capture on rising edge
		break;

	case eLow: // end of the pulldown phase of a bit
		gRcv.state = eHigh;
		TCCR1B &= ~_BV(ICES1); // capture on falling edge

		// check the low time to determine bit value
		if (ICR1L < MIN_ZERO_TIME)
			outchar = '<';
		else if (ICR1L <= MAX_ZERO_TIME)
			outchar = '0';
		else if (ICR1L < MIN_ONE_TIME)
			outchar = '=';
		else if (ICR1L <= MAX_ONE_TIME)
			outchar = '1';
		else
			outchar = '>';

		// hex conversion & output, for convenience
		gRcv.a4bits[gRcv.uBits % 4] = outchar;
		gRcv.uBits++;
		if ((gRcv.uBits % 4) == 0) // 4 bits completed?
		{
			uint8_t i;
			uint8_t uHexDigit = 0;
			
			for (i=0; i<4; i++) // convert to binary
			{
				if (gRcv.a4bits[i] == '1')
					uHexDigit |= (1 << (3-i));
				else if (gRcv.a4bits[i] != '0')
					uHexDigit = 0xFF; // mark error
			}
			MbusRxWriteByte(uHexDigit);
		}

		break;
	}
}

// timeout interrupt, this terminates a received packet
ISR(TIMER1_COMPA_vect)
{
	if (gRcv.state == eWait)
		return; // timeouts don't matter

	gRcv.state = eWait; // start looking for a new packet
	TCCR1B &= ~_BV(ICES1); // capture on falling edge

	// else the packet is completed
	if ((gRcv.uBits % 4) != 0) // there should be no data waiting for output
		MbusRxWriteByte(0xFE);


	MbusRxWriteByte(0xF0);
}

volatile static uint8_t half_sec_timeout = 14;
volatile static uint8_t one_sec_timeout = 28;
volatile static uint16_t power_off_timeout = 1687;

ISR(TIMER2_OVF_vect)
{
	half_sec_timeout--;
	if (!half_sec_timeout) 
	{
		half_sec_flag = 0xFF;	// 0.5 sec flag
		half_sec_timeout = 14;
	}

	one_sec_timeout--;
	if (!one_sec_timeout) 
	{
		one_sec_flag = 0xFF;	// 1 sec flag
		one_sec_timeout = 28;
	}

	if (PIND & _BV(PIN_POWER_ON))
	{
		power_off_timeout = 1687;
		power_off_timeout_flag = 0;
		power_off_flag = 0;

	}
	else
	{
		power_off_timeout--;
		if (!power_off_timeout) power_off_timeout_flag = 0xFF;
		if (power_off_timeout == 1673) power_off_flag = 0xFF;
	}	// power off timeout (1 min) flag

}

ISR(TWI_vect)
{
	if ((TWSR & 0xF8) == 0x08)
	{
		TWDR = i2c_addr;
		TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWIE);
	}
	else if ((TWSR & 0xF8) == 0x18)
	{
		TWDR = i2c_data;
		TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWIE);
	}
	else if ((TWSR & 0xF8) == 0x30)
	{
		TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO) | (1<<TWIE);
		twi_in_use = 0;
	}
	
	else if ((TWSR & 0xF8) == 0x40)
	{
		TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWIE);
	}
	else if ((TWSR & 0xF8) == 0x58)
	{
		i2c_data = TWDR;
		TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO) | (1<<TWIE);
		twi_in_use = 0;
	}
	else
	{
		TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO) | (1<<TWIE);
		twi_in_use = 0;
	}
}
